#include "libmaple.h"
#include "arduino/WProgram.h"

void setup(void);
void loop(void);
bool maple_transact(void);

struct maplepacket {
	unsigned char data_len; /* Bytes: header, data, and checksum */
	unsigned short data_len_rx; /* For Rx -- didn't realise we could get up to 512 bytes */
	unsigned char header[4];
	unsigned char data[600]; /* Our maximum packet size */
} packet;

void setup() {
	Serial.begin(57600); //the connection back to the computer.

	pinMode(13, OUTPUT); // LED pin on Arduino. Will light when accel/decel'ing

	// Maple bus data pins
	pinMode(4, INPUT);
	digitalWrite(4, HIGH);
	pinMode(5, INPUT);
	digitalWrite(5, HIGH);
	
	//I will be using Pin 7 for PWM output to my railway (speed).
	pinMode(7, OUTPUT);
	analogWrite(7, 0);
	//44 and 46 are the digital direction setting pins for my railway
	//one must be high, the other low, to work. reverse high/low for other direction.
	pinMode(44, OUTPUT);
	pinMode(46, OUTPUT);
	digitalWrite(46, HIGH);
	digitalWrite(44, LOW);

	//here is a quick packet to request device information from controller 'one'.
	//i.e. the only controller connected.
	packet.header[0] = 0; // Number of additional words in frame
	packet.header[1] = 0; // Sender address = Dreamcast
	packet.header[2] = (1 << 5); //(1 << 5); // Recipient address = main peripheral on port 0
	packet.header[3] = 1; // Command = request device information
	packet.data[0]   = 0x21; //checksum
	packet.data_len  = 5;
	//send the above packet to initiate comms with the controller.
	maple_transact();
}

unsigned char compute_checksum(unsigned char data_bytes)
{
	unsigned char *ptr = &(packet.header[0]);
	int count;
	unsigned char checksum = 0;

	for(count = 0; count < data_bytes + 4; count ++) {
		checksum ^= *ptr;
		ptr++;
	}
	return checksum;
}

bool maple_transact() {
	unsigned char *rx_buf_end;
	maple_tx_raw(&(packet.header[0]), packet.data_len);
	rx_buf_end = maple_rx_raw(&(packet.header[0]));
	packet.data_len_rx = (rx_buf_end - (&(packet.header[0])));
	return true;
}

/* Read packet to send from the controller. */
bool read_packet(void) {
	/* First byte: #bytes in packet (including header and checksum)*/
	while(!(Serial.available()))
		;
	packet.data_len = Serial.read();
	if(packet.data_len > 0) {
		unsigned char *data = &(packet.header[0]);
		for(int i = 0; i < packet.data_len; i++) {
			while(!(Serial.available()))
				//do nothing, we just want to wait for all bytes to come through;
			*data = Serial.read();
			data ++;
		}
		return true;
	} else {
		return false;
	}
}

void send_packet(void) {
	//send data back to the computer
	Serial.write((packet.data_len_rx & 0xff00) >> 8);
	Serial.write(packet.data_len_rx & 0xff);
	if(packet.data_len_rx) {
		Serial.write(&(packet.header[0]), packet.data_len_rx);
	}
}

int speed = 0;
int t_pos = 0;
int target_speed = 0;

void control_throttle(void) { 
	//quick hack, should actually do a binary 'and'
	//packet data[6] contains the throttle position, so use this to gauge
	//voltage output to pin 7 (or speed)
	switch (packet.data[6]) {
		case 0xF9: target_speed = 0;	t_pos = 0; break;
		case 0xFA: target_speed = 88;	t_pos = 1; break;
		case 0xFB: target_speed = 96;	t_pos = 2; break;
		case 0xFC: target_speed = 140;	t_pos = 3; break;
		case 0xFD: target_speed = 190;	t_pos = 4; break;
		case 0xFE: target_speed = 250;	t_pos = 5; break;
	}
}

void create_read_controller_packet() {
	//01 00 20 09 01 00 00 00 29
	//the above line is the data required to request the 'condition' of the controller
	packet.header[0] = 0x01; // Number of additional words in frame
	packet.header[1] = 0x00; // Sender address = Dreamcast
	packet.header[2] = (1 << 5); // Recipient address = main peripheral on port 0
	packet.header[3] = 0x09; // Command = request device information
	packet.data[0]   = 0x01;
	packet.data[1]   = 0x00;
	packet.data[2]   = 0x00;
	packet.data[3]   = 0x00;
	packet.data[4]   = 0x29;
	packet.data_len  = 9;
}

int main(void) {
    init();
    setup();

    for (;;) {
		//speeds lower than 40-50 are useless.. engine just hummmms...
		if (t_pos != 0 && speed < 40) speed = 40;
		else if (t_pos == 0 && speed < 40) speed = 0;

		create_read_controller_packet();
		maple_transact();
		control_throttle();

		digitalWrite(13, LOW);
		if (target_speed > speed) {
			//accel
			speed += t_pos / 2;
		} else if (target_speed < speed) {
			//speed += abs(target_speed - speed);
			//decel
			speed--;
		} else {
			digitalWrite(13, HIGH);
		}

		//send the speed to the pin/railway
		analogWrite(7, speed);
	}
	
	return 0;
}

